@@ -14,9 +14,6 @@ use arkworks_gadgets::{
 		constraints::AnchorLeafGadget, AnchorLeaf, Private as LeafPrivate, Public as LeafPublic,
 	},
 	merkle_tree::Path,
-	set::membership::{
-		constraints::SetMembershipGadget, Private as SetMembershipPrivate, SetMembership,
-	},
 };
 use arkworks_utils::{
 	poseidon::PoseidonParameters,
@@ -29,9 +26,6 @@ pub type Leaf_x5<F> = AnchorLeaf<F, PoseidonCRH_x5_4<F>>;
 
 pub type LeafGadget_x5<F> = AnchorLeafGadget<F, PoseidonCRH_x5_4<F>, PoseidonCRH_x5_4Gadget<F>>;
 
-pub type TestSetMembership<F, const M: usize> = SetMembership<F, M>;
-pub type TestSetMembershipGadget<F, const M: usize> = SetMembershipGadget<F, M>;
-
 pub type Circuit_x5<F, const N: usize, const M: usize> = AnchorCircuit<
 	F,
 	PoseidonCRH_x5_4<F>,
@@ -122,7 +116,7 @@ pub fn setup_proof_x5_4<E: PairingEngine, R: RngCore + CryptoRng>(
 	nullifier_raw: Vec<u8>,
 	leaves_raw: Vec<Vec<u8>>,
 	index: u64,
-	roots: Vec<Vec<u8>>,
+	roots: [Vec<u8>; M],
 	recipient_raw: Vec<u8>,
 	relayer_raw: Vec<u8>,
 	commitment_raw: Vec<u8>,
@@ -130,22 +124,12 @@ pub fn setup_proof_x5_4<E: PairingEngine, R: RngCore + CryptoRng>(
 	refund: u128,
 	pk: Vec<u8>,
 	rng: &mut R,
-) -> Result<
-	(
-		Vec<u8>,
-		Vec<u8>,
-		Vec<u8>,
-		Vec<u8>,
-		Vec<Vec<u8>>,
-		Vec<Vec<u8>>,
-	),
-	Error,
-> {
+) -> Result<(Vec<u8>, Vec<u8>, Vec<u8>, Vec<Vec<u8>>, Vec<Vec<u8>>), Error> {
 	let params3 = setup_params_x5_3::<E::Fr>(curve);
 	let params4 = setup_params_x5_4::<E::Fr>(curve);
 	let prover = AnchorProverSetupBn254_30::new(params3, params4);
 
-	let (circuit, leaf_raw, nullifier_hash_raw, root_raw, roots_raw, public_inputs_raw) = prover
+	let (circuit, leaf_raw, nullifier_hash_raw, roots_raw, public_inputs_raw) = prover
 		.setup_circuit_with_privates_raw(
 			chain_id,
 			secret_raw,
@@ -166,7 +150,6 @@ pub fn setup_proof_x5_4<E: PairingEngine, R: RngCore + CryptoRng>(
 		proof,
 		leaf_raw,
 		nullifier_hash_raw,
-		root_raw,
 		roots_raw,
 		public_inputs_raw,
 	))
@@ -197,10 +180,6 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		Self { params3, params4 }
 	}
 
-	pub fn setup_set(root: &F, roots: &[F; M]) -> Result<SetMembershipPrivate<F, M>, Error> {
-		TestSetMembership::generate_secrets(root, roots)
-	}
-
 	pub fn setup_arbitrary_data(
 		recipient: F,
 		relayer: F,
@@ -216,7 +195,6 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		chain_id: F,
 		nullifier_hash: F,
 		roots: [F; M],
-		root: F,
 		recipient: F,
 		relayer: F,
 		fee: F,
@@ -225,7 +203,7 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 	) -> Vec<F> {
 		let mut pub_ins = vec![chain_id, nullifier_hash];
 		pub_ins.extend(roots.to_vec());
-		pub_ins.extend(vec![root, recipient, relayer, fee, refund, commitment]);
+		pub_ins.extend(vec![recipient, relayer, fee, refund, commitment]);
 
 		pub_ins
 	}
@@ -237,7 +215,6 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		F,      // Chain id
 		F,      // Nullifier Hash
 		Vec<F>, // Roots
-		F,      // Root
 		F,      // Recipient
 		F,      // Relayer
 		F,      // Fee
@@ -248,17 +225,15 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		let nullifier_hash = public_inputs[1];
 		let offset = 2 + M;
 		let roots = public_inputs[2..offset].to_vec();
-		let root = public_inputs[offset + 1];
-		let recipient = public_inputs[offset + 2];
-		let relayer = public_inputs[offset + 3];
-		let fee = public_inputs[offset + 4];
-		let refund = public_inputs[offset + 5];
-		let commitments = public_inputs[offset + 6];
+		let recipient = public_inputs[offset + 1];
+		let relayer = public_inputs[offset + 2];
+		let fee = public_inputs[offset + 3];
+		let refund = public_inputs[offset + 4];
+		let commitments = public_inputs[offset + 5];
 		(
 			chain_id,
 			nullifier_hash,
 			roots,
-			root,
 			recipient,
 			relayer,
 			fee,
@@ -312,50 +287,33 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		chain_id: F,
 		leaves: &[F],
 		index: u64,
-		roots: &[F], // only first M - 1 member will be used
+		roots: [F; M], // only first M - 1 member will be used
 		recipient: F,
 		relayer: F,
 		fee: F,
 		refund: F,
 		commitment: F,
 		rng: &mut R,
-	) -> Result<(Circuit_x5<F, N, M>, F, F, F, Vec<F>, Vec<F>), Error> {
+	) -> Result<(Circuit_x5<F, N, M>, F, F, Vec<F>, Vec<F>), Error> {
 		let arbitrary_input =
 			Self::setup_arbitrary_data(recipient, relayer, fee, refund, commitment);
 		let (leaf_private, leaf_public, leaf, nullifier_hash) = self.setup_leaf(chain_id, rng)?;
-		let mut leaves_new = leaves.to_vec();
-		leaves_new.push(leaf);
-		let (tree, path) = self.setup_tree_and_path(&leaves_new, index)?;
-		let root = tree.root().inner();
-		let mut roots_new: [F; M] = [F::default(); M];
-		roots_new[0] = root;
-		let size_to_copy = if roots.len() > (M - 1) {
-			M - 1
-		} else {
-			roots.len()
-		};
-		for i in 0..size_to_copy {
-			roots_new[i + 1] = roots[i];
-		}
-		let set_private_inputs = Self::setup_set(&root, &roots_new)?;
+		let (_, path) = self.setup_tree_and_path(&leaves, index)?;
 
 		let mc = Circuit_x5::new(
 			arbitrary_input.clone(),
 			leaf_private,
 			leaf_public,
-			set_private_inputs,
-			roots_new,
+			roots,
 			self.params4,
 			path,
-			root.clone(),
 			nullifier_hash,
 		);
 
 		let public_inputs = Self::construct_public_inputs(
 			chain_id,
 			nullifier_hash,
-			roots_new,
-			root,
+			roots,
 			recipient,
 			relayer,
 			fee,
@@ -363,14 +321,7 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 			commitment,
 		);
 
-		Ok((
-			mc,
-			leaf,
-			nullifier_hash,
-			root,
-			roots_new.to_vec(),
-			public_inputs,
-		))
+		Ok((mc, leaf, nullifier_hash, roots.to_vec(), public_inputs))
 	}
 
 	#[allow(clippy::too_many_arguments)]
@@ -381,48 +332,33 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		nullifier: F,
 		leaves: &[F],
 		index: u64,
-		roots: &[F], // only first M - 1 member will be used
+		roots: [F; M],
 		recipient: F,
 		relayer: F,
 		fee: F,
 		refund: F,
 		commitment: F,
-	) -> Result<(Circuit_x5<F, N, M>, F, F, F, Vec<F>, Vec<F>), Error> {
+	) -> Result<(Circuit_x5<F, N, M>, F, F, Vec<F>, Vec<F>), Error> {
 		let arbitrary_input =
 			Self::setup_arbitrary_data(recipient, relayer, fee, refund, commitment);
 		let (leaf_private, leaf_public, leaf, nullifier_hash) =
 			self.setup_leaf_with_privates(chain_id, secret, nullifier)?;
-		let (tree, path) = self.setup_tree_and_path(&leaves, index)?;
-		let root = tree.root().inner();
-		let mut roots_new: [F; M] = [F::default(); M];
-		roots_new[0] = root;
-		let size_to_copy = if roots.len() > (M - 1) {
-			M - 1
-		} else {
-			roots.len()
-		};
-		for i in 0..size_to_copy {
-			roots_new[i + 1] = roots[i];
-		}
-		let set_private_inputs = Self::setup_set(&root, &roots_new)?;
+		let (_, path) = self.setup_tree_and_path(&leaves, index)?;
 
 		let mc = Circuit_x5::new(
 			arbitrary_input.clone(),
 			leaf_private,
 			leaf_public,
-			set_private_inputs,
-			roots_new,
+			roots,
 			self.params4,
 			path,
-			root.clone(),
 			nullifier_hash,
 		);
 
 		let public_inputs = Self::construct_public_inputs(
 			chain_id,
 			nullifier_hash,
-			roots_new,
-			root,
+			roots,
 			recipient,
 			relayer,
 			fee,
@@ -430,23 +366,16 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 			commitment,
 		);
 
-		Ok((
-			mc,
-			leaf,
-			nullifier_hash,
-			root,
-			roots_new.to_vec(),
-			public_inputs,
-		))
+		Ok((mc, leaf, nullifier_hash, roots.to_vec(), public_inputs))
 	}
 
 	pub fn setup_random_circuit<R: Rng>(
 		self,
 		rng: &mut R,
-	) -> Result<(Circuit_x5<F, N, M>, F, F, F, Vec<F>, Vec<F>), Error> {
+	) -> Result<(Circuit_x5<F, N, M>, F, F, Vec<F>, Vec<F>), Error> {
 		let chain_id = F::rand(rng);
 
-		let roots = Vec::new();
+		let roots = [F::rand(rng); M];
 		let recipient = F::rand(rng);
 		let relayer = F::rand(rng);
 		let fee = F::rand(rng);
@@ -460,7 +389,7 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		let index = 0;
 
 		self.setup_circuit_with_privates(
-			chain_id, secret, nullifier, &leaves, index, &roots, recipient, relayer, fee, refund,
+			chain_id, secret, nullifier, &leaves, index, roots, recipient, relayer, fee, refund,
 			commitment,
 		)
 	}
@@ -472,7 +401,7 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 		nullifier: Vec<u8>,
 		leaves: Vec<Vec<u8>>,
 		index: u64,
-		roots: Vec<Vec<u8>>,
+		roots: [Vec<u8>; M],
 		recipient: Vec<u8>,
 		relayer: Vec<u8>,
 		commitment: Vec<u8>,
@@ -483,7 +412,6 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 			Circuit_x5<F, N, M>,
 			Vec<u8>,
 			Vec<u8>,
-			Vec<u8>,
 			Vec<Vec<u8>>,
 			Vec<Vec<u8>>,
 		),
@@ -496,34 +424,32 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 			.iter()
 			.map(|x| F::from_le_bytes_mod_order(x))
 			.collect();
-		let roots_f: Vec<F> = roots
-			.iter()
-			.map(|x| F::from_le_bytes_mod_order(&x))
-			.collect();
+		let mut roots_f: [F; M] = [F::default(); M];
+		for i in 0..M {
+			roots_f[i] = F::from_le_bytes_mod_order(&roots[i]);
+		}
 		let recipient_f = F::from_le_bytes_mod_order(&recipient);
 		let relayer_f = F::from_le_bytes_mod_order(&relayer);
 		let commitment_f = F::from_le_bytes_mod_order(&commitment);
 		let fee_f = F::from(fee);
 		let refund_f = F::from(refund);
 
-		let (mc, leaf, nullifier_hash, root, roots, public_inputs) = self
-			.setup_circuit_with_privates(
-				chain_id_f,
-				secret_f,
-				nullifier_f,
-				&leaves_f,
-				index,
-				&roots_f,
-				recipient_f,
-				relayer_f,
-				fee_f,
-				refund_f,
-				commitment_f,
-			)?;
+		let (mc, leaf, nullifier_hash, roots, public_inputs) = self.setup_circuit_with_privates(
+			chain_id_f,
+			secret_f,
+			nullifier_f,
+			&leaves_f,
+			index,
+			roots_f,
+			recipient_f,
+			relayer_f,
+			fee_f,
+			refund_f,
+			commitment_f,
+		)?;
 
 		let leaf_raw = leaf.into_repr().to_bytes_le();
 		let nullifier_hash_raw = nullifier_hash.into_repr().to_bytes_le();
-		let root_raw = root.into_repr().to_bytes_le();
 		let roots_raw = roots.iter().map(|x| x.into_repr().to_bytes_le()).collect();
 		let public_inputs_raw: Vec<Vec<u8>> = public_inputs
 			.iter()
@@ -534,7 +460,6 @@ impl<F: PrimeField, const M: usize, const N: usize> AnchorProverSetup<F, M, N> {
 			mc,
 			leaf_raw,
 			nullifier_hash_raw,
-			root_raw,
 			roots_raw,
 			public_inputs_raw,
 		))
