@@ -6,17 +6,9 @@ use crate::{
 	},
 };
 use ark_bn254::Fr as Bn254Fr;
-use ark_crypto_primitives::SNARK;
-use ark_ec::PairingEngine;
+use ark_crypto_primitives::Error;
 use ark_ff::{BigInteger, PrimeField};
-use ark_groth16::{Groth16, Proof, ProvingKey, VerifyingKey};
-use ark_serialize::{CanonicalDeserialize, CanonicalSerialize};
-use ark_std::{
-	convert::TryInto,
-	rand::{CryptoRng, RngCore},
-	rc::Rc,
-	vec::Vec,
-};
+use ark_std::{rand::RngCore, rc::Rc, vec::Vec};
 use arkworks_gadgets::{
 	arbitrary::vanchor_data::VAnchorArbitraryData,
 	keypair::vanchor::Keypair,
@@ -28,8 +20,7 @@ use arkworks_utils::{
 	poseidon::PoseidonParameters,
 	utils::{
 		common::{
-			setup_params_x5_2, setup_params_x5_3, setup_params_x5_4, setup_params_x5_5,
-			verify_groth16, Curve,
+			setup_params_x5_2, setup_params_x5_3, setup_params_x5_4, setup_params_x5_5, Curve,
 		},
 		keccak_256, ExtData,
 	},
@@ -102,13 +93,13 @@ impl<
 		chain_ids: [u128; N],
 		amounts: [u128; N],
 		rng: &mut R,
-	) -> Vec<Utxo<F>> {
+	) -> Result<Vec<Utxo<F>>, Error> {
 		let chain_ids_f = chain_ids.map(|x| F::from(x));
 		let amounts_f = amounts.map(|x| F::from(x));
 
 		let keypairs = Self::setup_keypairs::<_, N>(rng);
 		let (commitments, nullifiers, leaf_privates, leaf_publics) =
-			self.setup_leaves(&chain_ids_f, &amounts_f, &keypairs, rng);
+			self.setup_leaves(&chain_ids_f, &amounts_f, &keypairs, rng)?;
 
 		let utxos: Vec<Utxo<F>> = (0..N)
 			.map(|i| Utxo {
@@ -121,7 +112,8 @@ impl<
 				commitment: commitments[i],
 			})
 			.collect();
-		utxos
+
+		Ok(utxos)
 	}
 
 	pub fn new_utxos_f<R: RngCore, const N: usize>(
@@ -129,10 +121,10 @@ impl<
 		chain_ids_f: [F; N],
 		amounts_f: [F; N],
 		rng: &mut R,
-	) -> Vec<Utxo<F>> {
+	) -> Result<Vec<Utxo<F>>, Error> {
 		let keypairs = Self::setup_keypairs::<_, N>(rng);
 		let (commitments, nullifiers, leaf_privates, leaf_publics) =
-			self.setup_leaves(&chain_ids_f, &amounts_f, &keypairs, rng);
+			self.setup_leaves(&chain_ids_f, &amounts_f, &keypairs, rng)?;
 
 		let utxos: Vec<Utxo<F>> = (0..N)
 			.map(|i| Utxo {
@@ -145,23 +137,27 @@ impl<
 				commitment: commitments[i],
 			})
 			.collect();
-		utxos
+
+		Ok(utxos)
 	}
 
 	pub fn setup_random_circuit<R: RngCore>(
 		self,
 		rng: &mut R,
-	) -> VACircuit<
-		F,
-		PoseidonCRH_x5_2<F>,
-		PoseidonCRH_x5_2Gadget<F>,
-		TreeConfig_x5<F>,
-		LeafCRHGadget<F>,
-		PoseidonCRH_x5_3Gadget<F>,
-		TREE_DEPTH,
-		INS,
-		OUTS,
-		M,
+	) -> Result<
+		VACircuit<
+			F,
+			PoseidonCRH_x5_2<F>,
+			PoseidonCRH_x5_2Gadget<F>,
+			TreeConfig_x5<F>,
+			LeafCRHGadget<F>,
+			PoseidonCRH_x5_3Gadget<F>,
+			TREE_DEPTH,
+			INS,
+			OUTS,
+			M,
+		>,
+		Error,
 	> {
 		let public_amount = rng.next_u64() as i128;
 
@@ -190,9 +186,9 @@ impl<
 			out_chain_ids,
 			out_amounts,
 			rng,
-		);
+		)?;
 
-		circuit
+		Ok(circuit)
 	}
 
 	pub fn setup_circuit_with_utxos(
@@ -207,26 +203,30 @@ impl<
 		in_utxos: Vec<Utxo<F>>,
 		// Output transactions
 		out_utxos: Vec<Utxo<F>>,
-	) -> (
-		VACircuit<
-			F,
-			PoseidonCRH_x5_2<F>,
-			PoseidonCRH_x5_2Gadget<F>,
-			TreeConfig_x5<F>,
-			LeafCRHGadget<F>,
-			PoseidonCRH_x5_3Gadget<F>,
-			TREE_DEPTH,
-			INS,
-			OUTS,
-			M,
-		>,
-		Vec<F>,
-	) {
+	) -> Result<
+		(
+			VACircuit<
+				F,
+				PoseidonCRH_x5_2<F>,
+				PoseidonCRH_x5_2Gadget<F>,
+				TreeConfig_x5<F>,
+				LeafCRHGadget<F>,
+				PoseidonCRH_x5_3Gadget<F>,
+				TREE_DEPTH,
+				INS,
+				OUTS,
+				M,
+			>,
+			Vec<F>,
+		),
+		Error,
+	> {
 		assert_eq!(in_utxos.len(), INS);
 		assert_eq!(out_utxos.len(), OUTS);
 		// Tree + set for proving input txos
+		let in_commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>();
 		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			self.setup_tree_and_set(&in_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>()[..]);
+			self.setup_tree_and_set(&in_commitments[..])?;
 
 		let ext_data = ExtData::new(
 			recipient,
@@ -255,18 +255,20 @@ impl<
 			in_set_private_inputs,
 			in_root_set,
 			out_utxos.clone(),
-		);
+		)?;
 
+		let in_nullifiers = in_utxos.iter().map(|x| x.nullifier).collect::<Vec<F>>();
+		let out_nullifiers = out_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>();
 		let public_inputs = Self::construct_public_inputs(
 			in_utxos[0].leaf_public.chain_id,
 			public_amount_f,
 			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<F>>(),
-			out_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>(),
+			in_nullifiers,
+			out_nullifiers,
 			ext_data_hash_f,
 		);
 
-		(circuit, public_inputs)
+		Ok((circuit, public_inputs))
 	}
 
 	// This function is used only for first transaction, when the tree is empty
@@ -282,31 +284,34 @@ impl<
 		out_chain_ids: [u128; OUTS],
 		out_amounts: [u128; OUTS],
 		rng: &mut R,
-	) -> (
-		VACircuit<
-			F,
-			PoseidonCRH_x5_2<F>,
-			PoseidonCRH_x5_2Gadget<F>,
-			TreeConfig_x5<F>,
-			LeafCRHGadget<F>,
-			PoseidonCRH_x5_3Gadget<F>,
-			TREE_DEPTH,
-			INS,
-			OUTS,
-			M,
-		>,
-		Vec<F>,
-		Vec<Utxo<F>>,
-		Vec<Utxo<F>>,
-	) {
+	) -> Result<
+		(
+			VACircuit<
+				F,
+				PoseidonCRH_x5_2<F>,
+				PoseidonCRH_x5_2Gadget<F>,
+				TreeConfig_x5<F>,
+				LeafCRHGadget<F>,
+				PoseidonCRH_x5_3Gadget<F>,
+				TREE_DEPTH,
+				INS,
+				OUTS,
+				M,
+			>,
+			Vec<F>,
+			Vec<Utxo<F>>,
+			Vec<Utxo<F>>,
+		),
+		Error,
+	> {
 		// Making a vec of same chain ids to be passed into setup_leaves
 		let in_chain_ids = [in_chain_id; INS];
 
 		// Input leaves (txos)
-		let in_utxos = self.new_utxos(in_chain_ids, in_amounts, rng);
+		let in_utxos = self.new_utxos(in_chain_ids, in_amounts, rng)?;
 
 		// Output leaves (txos)
-		let out_utxos = self.new_utxos(out_chain_ids, out_amounts, rng);
+		let out_utxos = self.new_utxos(out_chain_ids, out_amounts, rng)?;
 
 		let (circuit, public_inputs) = self.setup_circuit_with_utxos(
 			public_amount,
@@ -316,9 +321,9 @@ impl<
 			fee,
 			in_utxos.clone(),
 			out_utxos.clone(),
-		);
+		)?;
 
-		(circuit, public_inputs, in_utxos, out_utxos)
+		Ok((circuit, public_inputs, in_utxos, out_utxos))
 	}
 
 	pub fn setup_circuit(
@@ -334,24 +339,44 @@ impl<
 		in_root_set: [F; M],
 		// Output transactions
 		out_utxos: Vec<Utxo<F>>,
-	) -> VACircuit<
-		F,
-		PoseidonCRH_x5_2<F>,
-		PoseidonCRH_x5_2Gadget<F>,
-		TreeConfig_x5<F>,
-		LeafCRHGadget<F>,
-		PoseidonCRH_x5_3Gadget<F>,
-		TREE_DEPTH,
-		INS,
-		OUTS,
-		M,
+	) -> Result<
+		VACircuit<
+			F,
+			PoseidonCRH_x5_2<F>,
+			PoseidonCRH_x5_2Gadget<F>,
+			TreeConfig_x5<F>,
+			LeafCRHGadget<F>,
+			PoseidonCRH_x5_3Gadget<F>,
+			TREE_DEPTH,
+			INS,
+			OUTS,
+			M,
+		>,
+		Error,
 	> {
-		let out_pub_keys = out_utxos
+		let in_leaf_private_inputs = in_utxos
+			.iter()
+			.map(|x| x.leaf_private.clone())
+			.collect::<Vec<LeafPrivateInput<F>>>();
+		let in_keypair_inputs = in_utxos
 			.iter()
 			.map(|x| x.keypair.clone())
-			.map(|x| x.public_key(&self.params2).unwrap())
-			.collect();
+			.collect::<Vec<Keypair<F, PoseidonCRH_x5_2<F>>>>();
+		let in_nullifiers = in_utxos.iter().map(|x| x.nullifier).collect::<Vec<F>>();
 
+		let out_pub_keys: Result<Vec<F>, _> = out_utxos
+			.iter()
+			.map(|x| x.keypair.public_key(&self.params2))
+			.collect();
+		let out_commitments = out_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>();
+		let out_leaf_private = out_utxos
+			.iter()
+			.map(|x| x.leaf_private.clone())
+			.collect::<Vec<LeafPrivateInput<F>>>();
+		let out_leaf_public = out_utxos
+			.iter()
+			.map(|x| x.leaf_public.clone())
+			.collect::<Vec<LeafPublicInput<F>>>();
 		let circuit = VACircuit::<
 			F,
 			PoseidonCRH_x5_2<F>,
@@ -366,14 +391,8 @@ impl<
 		>::new(
 			public_amount,
 			arbitrary_data,
-			in_utxos
-				.iter()
-				.map(|x| x.leaf_private.clone())
-				.collect::<Vec<LeafPrivateInput<F>>>(),
-			in_utxos
-				.iter()
-				.map(|x| x.keypair.clone())
-				.collect::<Vec<Keypair<F, PoseidonCRH_x5_2<F>>>>(),
+			in_leaf_private_inputs,
+			in_keypair_inputs,
 			in_utxos[0].leaf_public.clone(),
 			in_set_private_inputs,
 			in_root_set,
@@ -382,75 +401,14 @@ impl<
 			self.params5,
 			in_paths,
 			in_indicies,
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<F>>(),
-			out_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.leaf_private.clone())
-				.collect::<Vec<LeafPrivateInput<F>>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.leaf_public.clone())
-				.collect::<Vec<LeafPublicInput<F>>>(),
-			out_pub_keys,
+			in_nullifiers,
+			out_commitments,
+			out_leaf_private,
+			out_leaf_public,
+			out_pub_keys?,
 		);
 
-		circuit
-	}
-
-	pub fn setup_keys<E: PairingEngine, R: RngCore + CryptoRng>(
-		circuit: VACircuit<
-			E::Fr,
-			PoseidonCRH_x5_2<E::Fr>,
-			PoseidonCRH_x5_2Gadget<E::Fr>,
-			TreeConfig_x5<E::Fr>,
-			LeafCRHGadget<E::Fr>,
-			PoseidonCRH_x5_3Gadget<E::Fr>,
-			TREE_DEPTH,
-			INS,
-			OUTS,
-			M,
-		>,
-		rng: &mut R,
-	) -> (Vec<u8>, Vec<u8>) {
-		let (pk, vk) = Groth16::<E>::circuit_specific_setup(circuit, rng).unwrap();
-
-		let mut pk_bytes = Vec::new();
-		let mut vk_bytes = Vec::new();
-		pk.serialize(&mut pk_bytes).unwrap();
-		vk.serialize(&mut vk_bytes).unwrap();
-		(pk_bytes, vk_bytes)
-	}
-
-	pub fn prove<E: PairingEngine, R: RngCore + CryptoRng>(
-		circuit: VACircuit<
-			E::Fr,
-			PoseidonCRH_x5_2<E::Fr>,
-			PoseidonCRH_x5_2Gadget<E::Fr>,
-			TreeConfig_x5<E::Fr>,
-			LeafCRHGadget<E::Fr>,
-			PoseidonCRH_x5_3Gadget<E::Fr>,
-			TREE_DEPTH,
-			INS,
-			OUTS,
-			M,
-		>,
-		pk_bytes: &[u8],
-		rng: &mut R,
-	) -> Vec<u8> {
-		let pk = ProvingKey::<E>::deserialize(pk_bytes).unwrap();
-
-		let proof = Groth16::prove(&pk, circuit, rng).unwrap();
-		let mut proof_bytes = Vec::new();
-		proof.serialize(&mut proof_bytes).unwrap();
-		proof_bytes
-	}
-
-	pub fn verify<E: PairingEngine>(public_inputs: &Vec<E::Fr>, vk: &[u8], proof: &[u8]) -> bool {
-		let vk = VerifyingKey::<E>::deserialize(vk).unwrap();
-		let proof = Proof::<E>::deserialize(proof).unwrap();
-		let ver_res = verify_groth16(&vk, &public_inputs, &proof);
-		ver_res
+		Ok(circuit)
 	}
 
 	pub fn setup_keypairs<R: RngCore, const N: usize>(
@@ -465,18 +423,21 @@ impl<
 		amounts: &[F; N],
 		keypairs: &[Keypair<F, PoseidonCRH_x5_2<F>>; N],
 		rng: &mut R,
-	) -> (
-		[F; N],
-		[F; N],
-		[LeafPrivateInput<F>; N],
-		[LeafPublicInput<F>; N],
-	) {
+	) -> Result<
+		(
+			[F; N],
+			[F; N],
+			[LeafPrivateInput<F>; N],
+			[LeafPublicInput<F>; N],
+		),
+		Error,
+	> {
 		let num_inputs = amounts.len();
 
-		let mut leaves = Vec::new();
-		let mut nullifiers = Vec::new();
-		let mut private_inputs = Vec::new();
-		let mut public_inputs = Vec::new();
+		let mut leaves = [F::default(); N];
+		let mut nullifiers = [F::default(); N];
+		let mut private_inputs = [LeafPrivateInput::<F>::default(); N];
+		let mut public_inputs = [LeafPublicInput::<F>::default(); N];
 
 		for i in 0..num_inputs {
 			let chain_id = F::from(chain_ids[i]);
@@ -487,43 +448,39 @@ impl<
 			let private_input = LeafPrivateInput::<F>::new(amount, blinding);
 			let public_input = LeafPublicInput::<F>::new(chain_id);
 
-			let pub_key = keypairs[i].public_key(&self.params2).unwrap();
+			let pub_key = keypairs[i].public_key(&self.params2)?;
 
 			let leaf = Leaf::<F, PoseidonCRH_x5_4<F>>::create_leaf(
 				&private_input,
 				&public_input,
 				&pub_key,
 				&self.params5,
-			)
-			.unwrap();
+			)?;
 
-			let signature = keypairs[i].signature(&leaf, &index, &self.params4).unwrap();
+			let signature = keypairs[i].signature(&leaf, &index, &self.params4)?;
 
-			let nullfier = Leaf::<F, PoseidonCRH_x5_4<F>>::create_nullifier(
+			let nullifier = Leaf::<F, PoseidonCRH_x5_4<F>>::create_nullifier(
 				&signature,
 				&leaf,
 				&self.params4,
 				&index,
-			)
-			.unwrap();
+			)?;
 
-			leaves.push(leaf);
-			nullifiers.push(nullfier);
-			private_inputs.push(private_input);
-			public_inputs.push(public_input);
+			leaves[i] = leaf;
+			nullifiers[i] = nullifier;
+			private_inputs[i] = private_input;
+			public_inputs[i] = public_input;
 		}
 
-		(
-			leaves.try_into().unwrap(),
-			nullifiers.try_into().unwrap(),
-			private_inputs.try_into().unwrap(),
-			public_inputs.try_into().unwrap(),
-		)
+		Ok((leaves, nullifiers, private_inputs, public_inputs))
 	}
 
-	pub fn setup_tree(&self, leaves: &[F]) -> (Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>, Vec<F>, F) {
+	pub fn setup_tree(
+		&self,
+		leaves: &[F],
+	) -> Result<(Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>, Vec<F>, F), Error> {
 		let inner_params = Rc::new(self.params3.clone());
-		let tree = Tree_x5::new_sequential(inner_params, Rc::new(()), &leaves.to_vec()).unwrap();
+		let tree = Tree_x5::new_sequential(inner_params, Rc::new(()), &leaves.to_vec())?;
 		let root = tree.root();
 
 		let num_leaves = leaves.len();
@@ -532,38 +489,41 @@ impl<
 		let mut indices = Vec::new();
 		for i in 0..num_leaves {
 			let path = tree.generate_membership_proof::<TREE_DEPTH>(i as u64);
-			let index = path.get_index(&root, &leaves[i]).unwrap();
+			let index = path.get_index(&root, &leaves[i])?;
 			paths.push(path);
 			indices.push(index);
 		}
 
-		(paths, indices, root.inner())
+		Ok((paths, indices, root.inner()))
 	}
 
-	pub fn setup_root_set(root: F) -> ([F; M], Vec<SetPrivateInputs<F, M>>) {
+	pub fn setup_root_set(root: F) -> Result<([F; M], Vec<SetPrivateInputs<F, M>>), Error> {
 		let root_set = [root.clone(); M];
 
 		let mut set_private_inputs = Vec::new();
 		for _ in 0..M {
-			let set_private_input = SetMembership::generate_secrets(&root, &root_set).unwrap();
+			let set_private_input = SetMembership::generate_secrets(&root, &root_set)?;
 			set_private_inputs.push(set_private_input);
 		}
 
-		(root_set, set_private_inputs)
+		Ok((root_set, set_private_inputs))
 	}
 
 	pub fn setup_tree_and_set(
 		&self,
 		leaves: &[F],
-	) -> (
-		Vec<F>,
-		Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>,
-		Vec<SetPrivateInputs<F, M>>,
-		[F; M],
-	) {
-		let (paths, indices, root) = self.setup_tree(&leaves);
-		let (root_set, set_private_inputs) = Self::setup_root_set(root);
-		(indices, paths, set_private_inputs, root_set)
+	) -> Result<
+		(
+			Vec<F>,
+			Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>,
+			Vec<SetPrivateInputs<F, M>>,
+			[F; M],
+		),
+		Error,
+	> {
+		let (paths, indices, root) = self.setup_tree(&leaves)?;
+		let (root_set, set_private_inputs) = Self::setup_root_set(root)?;
+		Ok((indices, paths, set_private_inputs, root_set))
 	}
 
 	pub fn construct_public_inputs(
@@ -583,30 +543,52 @@ impl<
 		public_inputs
 	}
 
+	// TODO: Fix to match INS and OUTS
 	pub fn deconstruct_public_inputs(
-		public_inputs: &Vec<F>,
-	) -> (
-		F,      // Chain Id
-		F,      // Public amount
-		Vec<F>, // Roots
-		Vec<F>, // Input tx Nullifiers
-		Vec<F>, // Output tx commitments
-		F,      // External data hash
-	) {
-		let public_amount = public_inputs[0];
-		let ext_data_hash = public_inputs[1];
-		let nullifiers = public_inputs[2..4].to_vec();
-		let commitments = public_inputs[4..6].to_vec();
-		let chain_id = public_inputs[6];
-		let root_set = public_inputs[7..9].to_vec();
+		public_inputs: Vec<F>,
+	) -> Result<
 		(
+			F,      // Chain Id
+			F,      // Public amount
+			Vec<F>, // Roots
+			Vec<F>, // Input tx Nullifiers
+			Vec<F>, // Output tx commitments
+			F,      // External data hash
+		),
+		Error,
+	> {
+		let mut pub_ins = public_inputs;
+
+		let err = "Invalid public inputs";
+
+		let mut root_set = Vec::new();
+		for _ in 0..INS {
+			root_set.push(pub_ins.pop().ok_or(err)?);
+		}
+
+		let chain_id = pub_ins.pop().ok_or(err)?;
+
+		let mut out_commitments = Vec::new();
+		for _ in 0..OUTS {
+			out_commitments.push(pub_ins.pop().ok_or(err)?);
+		}
+
+		let mut in_nullifiers = Vec::new();
+		for _ in 0..INS {
+			in_nullifiers.push(pub_ins.pop().ok_or(err)?);
+		}
+
+		let ext_data_hash = pub_ins.pop().ok_or(err)?;
+		let public_amount = pub_ins.pop().ok_or(err)?;
+
+		Ok((
 			chain_id,
 			public_amount,
 			root_set,
-			nullifiers,
-			commitments,
+			in_nullifiers,
+			out_commitments,
 			ext_data_hash,
-		)
+		))
 	}
 
 	pub fn setup_arbitrary_data(ext_data: F) -> VAnchorArbitraryData<F> {
